###
# Copyright CEA (Commissariat à l'énergie atomique et aux
# énergies alternatives) (2017-2025)
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
###
import copy
import random
from typing import List, Optional, Tuple, Union

import numpy as np

from ..models import hamming_weight

# fmt: off
AES_SBOX_TABLE = np.array([
    0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5,
    0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
    0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0,
    0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
    0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc,
    0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
    0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a,
    0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
    0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0,
    0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
    0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b,
    0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
    0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85,
    0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
    0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5,
    0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
    0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17,
    0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
    0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88,
    0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
    0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c,
    0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
    0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9,
    0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
    0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6,
    0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
    0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e,
    0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
    0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94,
    0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
    0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68,
    0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16], dtype=np.uint8)

AES_INV_SBOX_TABLE = np.array([
    0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e,
    0x81, 0xf3, 0xd7, 0xfb, 0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87,
    0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb, 0x54, 0x7b, 0x94, 0x32,
    0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e,
    0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49,
    0x6d, 0x8b, 0xd1, 0x25, 0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16,
    0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92, 0x6c, 0x70, 0x48, 0x50,
    0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84,
    0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05,
    0xb8, 0xb3, 0x45, 0x06, 0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02,
    0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b, 0x3a, 0x91, 0x11, 0x41,
    0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73,
    0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8,
    0x1c, 0x75, 0xdf, 0x6e, 0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89,
    0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b, 0xfc, 0x56, 0x3e, 0x4b,
    0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4,
    0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59,
    0x27, 0x80, 0xec, 0x5f, 0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d,
    0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef, 0xa0, 0xe0, 0x3b, 0x4d,
    0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61,
    0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63,
    0x55, 0x21, 0x0c, 0x7d], dtype=np.uint8)

AES_TE0 = np.array([
    0xc66363a5, 0xf87c7c84, 0xee777799, 0xf67b7b8d,
    0xfff2f20d, 0xd66b6bbd, 0xde6f6fb1, 0x91c5c554,
    0x60303050, 0x02010103, 0xce6767a9, 0x562b2b7d,
    0xe7fefe19, 0xb5d7d762, 0x4dababe6, 0xec76769a,
    0x8fcaca45, 0x1f82829d, 0x89c9c940, 0xfa7d7d87,
    0xeffafa15, 0xb25959eb, 0x8e4747c9, 0xfbf0f00b,
    0x41adadec, 0xb3d4d467, 0x5fa2a2fd, 0x45afafea,
    0x239c9cbf, 0x53a4a4f7, 0xe4727296, 0x9bc0c05b,
    0x75b7b7c2, 0xe1fdfd1c, 0x3d9393ae, 0x4c26266a,
    0x6c36365a, 0x7e3f3f41, 0xf5f7f702, 0x83cccc4f,
    0x6834345c, 0x51a5a5f4, 0xd1e5e534, 0xf9f1f108,
    0xe2717193, 0xabd8d873, 0x62313153, 0x2a15153f,
    0x0804040c, 0x95c7c752, 0x46232365, 0x9dc3c35e,
    0x30181828, 0x379696a1, 0x0a05050f, 0x2f9a9ab5,
    0x0e070709, 0x24121236, 0x1b80809b, 0xdfe2e23d,
    0xcdebeb26, 0x4e272769, 0x7fb2b2cd, 0xea75759f,
    0x1209091b, 0x1d83839e, 0x582c2c74, 0x341a1a2e,
    0x361b1b2d, 0xdc6e6eb2, 0xb45a5aee, 0x5ba0a0fb,
    0xa45252f6, 0x763b3b4d, 0xb7d6d661, 0x7db3b3ce,
    0x5229297b, 0xdde3e33e, 0x5e2f2f71, 0x13848497,
    0xa65353f5, 0xb9d1d168, 0x00000000, 0xc1eded2c,
    0x40202060, 0xe3fcfc1f, 0x79b1b1c8, 0xb65b5bed,
    0xd46a6abe, 0x8dcbcb46, 0x67bebed9, 0x7239394b,
    0x944a4ade, 0x984c4cd4, 0xb05858e8, 0x85cfcf4a,
    0xbbd0d06b, 0xc5efef2a, 0x4faaaae5, 0xedfbfb16,
    0x864343c5, 0x9a4d4dd7, 0x66333355, 0x11858594,
    0x8a4545cf, 0xe9f9f910, 0x04020206, 0xfe7f7f81,
    0xa05050f0, 0x783c3c44, 0x259f9fba, 0x4ba8a8e3,
    0xa25151f3, 0x5da3a3fe, 0x804040c0, 0x058f8f8a,
    0x3f9292ad, 0x219d9dbc, 0x70383848, 0xf1f5f504,
    0x63bcbcdf, 0x77b6b6c1, 0xafdada75, 0x42212163,
    0x20101030, 0xe5ffff1a, 0xfdf3f30e, 0xbfd2d26d,
    0x81cdcd4c, 0x180c0c14, 0x26131335, 0xc3ecec2f,
    0xbe5f5fe1, 0x359797a2, 0x884444cc, 0x2e171739,
    0x93c4c457, 0x55a7a7f2, 0xfc7e7e82, 0x7a3d3d47,
    0xc86464ac, 0xba5d5de7, 0x3219192b, 0xe6737395,
    0xc06060a0, 0x19818198, 0x9e4f4fd1, 0xa3dcdc7f,
    0x44222266, 0x542a2a7e, 0x3b9090ab, 0x0b888883,
    0x8c4646ca, 0xc7eeee29, 0x6bb8b8d3, 0x2814143c,
    0xa7dede79, 0xbc5e5ee2, 0x160b0b1d, 0xaddbdb76,
    0xdbe0e03b, 0x64323256, 0x743a3a4e, 0x140a0a1e,
    0x924949db, 0x0c06060a, 0x4824246c, 0xb85c5ce4,
    0x9fc2c25d, 0xbdd3d36e, 0x43acacef, 0xc46262a6,
    0x399191a8, 0x319595a4, 0xd3e4e437, 0xf279798b,
    0xd5e7e732, 0x8bc8c843, 0x6e373759, 0xda6d6db7,
    0x018d8d8c, 0xb1d5d564, 0x9c4e4ed2, 0x49a9a9e0,
    0xd86c6cb4, 0xac5656fa, 0xf3f4f407, 0xcfeaea25,
    0xca6565af, 0xf47a7a8e, 0x47aeaee9, 0x10080818,
    0x6fbabad5, 0xf0787888, 0x4a25256f, 0x5c2e2e72,
    0x381c1c24, 0x57a6a6f1, 0x73b4b4c7, 0x97c6c651,
    0xcbe8e823, 0xa1dddd7c, 0xe874749c, 0x3e1f1f21,
    0x964b4bdd, 0x61bdbddc, 0x0d8b8b86, 0x0f8a8a85,
    0xe0707090, 0x7c3e3e42, 0x71b5b5c4, 0xcc6666aa,
    0x904848d8, 0x06030305, 0xf7f6f601, 0x1c0e0e12,
    0xc26161a3, 0x6a35355f, 0xae5757f9, 0x69b9b9d0,
    0x17868691, 0x99c1c158, 0x3a1d1d27, 0x279e9eb9,
    0xd9e1e138, 0xebf8f813, 0x2b9898b3, 0x22111133,
    0xd26969bb, 0xa9d9d970, 0x078e8e89, 0x339494a7,
    0x2d9b9bb6, 0x3c1e1e22, 0x15878792, 0xc9e9e920,
    0x87cece49, 0xaa5555ff, 0x50282878, 0xa5dfdf7a,
    0x038c8c8f, 0x59a1a1f8, 0x09898980, 0x1a0d0d17,
    0x65bfbfda, 0xd7e6e631, 0x844242c6, 0xd06868b8,
    0x824141c3, 0x299999b0, 0x5a2d2d77, 0x1e0f0f11,
    0x7bb0b0cb, 0xa85454fc, 0x6dbbbbd6, 0x2c16163a], dtype=np.uint32)

AES_TE1 = np.array([
    0xa5c66363, 0x84f87c7c, 0x99ee7777, 0x8df67b7b,
    0x0dfff2f2, 0xbdd66b6b, 0xb1de6f6f, 0x5491c5c5,
    0x50603030, 0x03020101, 0xa9ce6767, 0x7d562b2b,
    0x19e7fefe, 0x62b5d7d7, 0xe64dabab, 0x9aec7676,
    0x458fcaca, 0x9d1f8282, 0x4089c9c9, 0x87fa7d7d,
    0x15effafa, 0xebb25959, 0xc98e4747, 0x0bfbf0f0,
    0xec41adad, 0x67b3d4d4, 0xfd5fa2a2, 0xea45afaf,
    0xbf239c9c, 0xf753a4a4, 0x96e47272, 0x5b9bc0c0,
    0xc275b7b7, 0x1ce1fdfd, 0xae3d9393, 0x6a4c2626,
    0x5a6c3636, 0x417e3f3f, 0x02f5f7f7, 0x4f83cccc,
    0x5c683434, 0xf451a5a5, 0x34d1e5e5, 0x08f9f1f1,
    0x93e27171, 0x73abd8d8, 0x53623131, 0x3f2a1515,
    0x0c080404, 0x5295c7c7, 0x65462323, 0x5e9dc3c3,
    0x28301818, 0xa1379696, 0x0f0a0505, 0xb52f9a9a,
    0x090e0707, 0x36241212, 0x9b1b8080, 0x3ddfe2e2,
    0x26cdebeb, 0x694e2727, 0xcd7fb2b2, 0x9fea7575,
    0x1b120909, 0x9e1d8383, 0x74582c2c, 0x2e341a1a,
    0x2d361b1b, 0xb2dc6e6e, 0xeeb45a5a, 0xfb5ba0a0,
    0xf6a45252, 0x4d763b3b, 0x61b7d6d6, 0xce7db3b3,
    0x7b522929, 0x3edde3e3, 0x715e2f2f, 0x97138484,
    0xf5a65353, 0x68b9d1d1, 0x00000000, 0x2cc1eded,
    0x60402020, 0x1fe3fcfc, 0xc879b1b1, 0xedb65b5b,
    0xbed46a6a, 0x468dcbcb, 0xd967bebe, 0x4b723939,
    0xde944a4a, 0xd4984c4c, 0xe8b05858, 0x4a85cfcf,
    0x6bbbd0d0, 0x2ac5efef, 0xe54faaaa, 0x16edfbfb,
    0xc5864343, 0xd79a4d4d, 0x55663333, 0x94118585,
    0xcf8a4545, 0x10e9f9f9, 0x06040202, 0x81fe7f7f,
    0xf0a05050, 0x44783c3c, 0xba259f9f, 0xe34ba8a8,
    0xf3a25151, 0xfe5da3a3, 0xc0804040, 0x8a058f8f,
    0xad3f9292, 0xbc219d9d, 0x48703838, 0x04f1f5f5,
    0xdf63bcbc, 0xc177b6b6, 0x75afdada, 0x63422121,
    0x30201010, 0x1ae5ffff, 0x0efdf3f3, 0x6dbfd2d2,
    0x4c81cdcd, 0x14180c0c, 0x35261313, 0x2fc3ecec,
    0xe1be5f5f, 0xa2359797, 0xcc884444, 0x392e1717,
    0x5793c4c4, 0xf255a7a7, 0x82fc7e7e, 0x477a3d3d,
    0xacc86464, 0xe7ba5d5d, 0x2b321919, 0x95e67373,
    0xa0c06060, 0x98198181, 0xd19e4f4f, 0x7fa3dcdc,
    0x66442222, 0x7e542a2a, 0xab3b9090, 0x830b8888,
    0xca8c4646, 0x29c7eeee, 0xd36bb8b8, 0x3c281414,
    0x79a7dede, 0xe2bc5e5e, 0x1d160b0b, 0x76addbdb,
    0x3bdbe0e0, 0x56643232, 0x4e743a3a, 0x1e140a0a,
    0xdb924949, 0x0a0c0606, 0x6c482424, 0xe4b85c5c,
    0x5d9fc2c2, 0x6ebdd3d3, 0xef43acac, 0xa6c46262,
    0xa8399191, 0xa4319595, 0x37d3e4e4, 0x8bf27979,
    0x32d5e7e7, 0x438bc8c8, 0x596e3737, 0xb7da6d6d,
    0x8c018d8d, 0x64b1d5d5, 0xd29c4e4e, 0xe049a9a9,
    0xb4d86c6c, 0xfaac5656, 0x07f3f4f4, 0x25cfeaea,
    0xafca6565, 0x8ef47a7a, 0xe947aeae, 0x18100808,
    0xd56fbaba, 0x88f07878, 0x6f4a2525, 0x725c2e2e,
    0x24381c1c, 0xf157a6a6, 0xc773b4b4, 0x5197c6c6,
    0x23cbe8e8, 0x7ca1dddd, 0x9ce87474, 0x213e1f1f,
    0xdd964b4b, 0xdc61bdbd, 0x860d8b8b, 0x850f8a8a,
    0x90e07070, 0x427c3e3e, 0xc471b5b5, 0xaacc6666,
    0xd8904848, 0x05060303, 0x01f7f6f6, 0x121c0e0e,
    0xa3c26161, 0x5f6a3535, 0xf9ae5757, 0xd069b9b9,
    0x91178686, 0x5899c1c1, 0x273a1d1d, 0xb9279e9e,
    0x38d9e1e1, 0x13ebf8f8, 0xb32b9898, 0x33221111,
    0xbbd26969, 0x70a9d9d9, 0x89078e8e, 0xa7339494,
    0xb62d9b9b, 0x223c1e1e, 0x92158787, 0x20c9e9e9,
    0x4987cece, 0xffaa5555, 0x78502828, 0x7aa5dfdf,
    0x8f038c8c, 0xf859a1a1, 0x80098989, 0x171a0d0d,
    0xda65bfbf, 0x31d7e6e6, 0xc6844242, 0xb8d06868,
    0xc3824141, 0xb0299999, 0x775a2d2d, 0x111e0f0f,
    0xcb7bb0b0, 0xfca85454, 0xd66dbbbb, 0x3a2c1616],
    dtype=np.uint32)

AES_TE2 = np.array([
    0x63a5c663, 0x7c84f87c, 0x7799ee77, 0x7b8df67b,
    0xf20dfff2, 0x6bbdd66b, 0x6fb1de6f, 0xc55491c5,
    0x30506030, 0x01030201, 0x67a9ce67, 0x2b7d562b,
    0xfe19e7fe, 0xd762b5d7, 0xabe64dab, 0x769aec76,
    0xca458fca, 0x829d1f82, 0xc94089c9, 0x7d87fa7d,
    0xfa15effa, 0x59ebb259, 0x47c98e47, 0xf00bfbf0,
    0xadec41ad, 0xd467b3d4, 0xa2fd5fa2, 0xafea45af,
    0x9cbf239c, 0xa4f753a4, 0x7296e472, 0xc05b9bc0,
    0xb7c275b7, 0xfd1ce1fd, 0x93ae3d93, 0x266a4c26,
    0x365a6c36, 0x3f417e3f, 0xf702f5f7, 0xcc4f83cc,
    0x345c6834, 0xa5f451a5, 0xe534d1e5, 0xf108f9f1,
    0x7193e271, 0xd873abd8, 0x31536231, 0x153f2a15,
    0x040c0804, 0xc75295c7, 0x23654623, 0xc35e9dc3,
    0x18283018, 0x96a13796, 0x050f0a05, 0x9ab52f9a,
    0x07090e07, 0x12362412, 0x809b1b80, 0xe23ddfe2,
    0xeb26cdeb, 0x27694e27, 0xb2cd7fb2, 0x759fea75,
    0x091b1209, 0x839e1d83, 0x2c74582c, 0x1a2e341a,
    0x1b2d361b, 0x6eb2dc6e, 0x5aeeb45a, 0xa0fb5ba0,
    0x52f6a452, 0x3b4d763b, 0xd661b7d6, 0xb3ce7db3,
    0x297b5229, 0xe33edde3, 0x2f715e2f, 0x84971384,
    0x53f5a653, 0xd168b9d1, 0x00000000, 0xed2cc1ed,
    0x20604020, 0xfc1fe3fc, 0xb1c879b1, 0x5bedb65b,
    0x6abed46a, 0xcb468dcb, 0xbed967be, 0x394b7239,
    0x4ade944a, 0x4cd4984c, 0x58e8b058, 0xcf4a85cf,
    0xd06bbbd0, 0xef2ac5ef, 0xaae54faa, 0xfb16edfb,
    0x43c58643, 0x4dd79a4d, 0x33556633, 0x85941185,
    0x45cf8a45, 0xf910e9f9, 0x02060402, 0x7f81fe7f,
    0x50f0a050, 0x3c44783c, 0x9fba259f, 0xa8e34ba8,
    0x51f3a251, 0xa3fe5da3, 0x40c08040, 0x8f8a058f,
    0x92ad3f92, 0x9dbc219d, 0x38487038, 0xf504f1f5,
    0xbcdf63bc, 0xb6c177b6, 0xda75afda, 0x21634221,
    0x10302010, 0xff1ae5ff, 0xf30efdf3, 0xd26dbfd2,
    0xcd4c81cd, 0x0c14180c, 0x13352613, 0xec2fc3ec,
    0x5fe1be5f, 0x97a23597, 0x44cc8844, 0x17392e17,
    0xc45793c4, 0xa7f255a7, 0x7e82fc7e, 0x3d477a3d,
    0x64acc864, 0x5de7ba5d, 0x192b3219, 0x7395e673,
    0x60a0c060, 0x81981981, 0x4fd19e4f, 0xdc7fa3dc,
    0x22664422, 0x2a7e542a, 0x90ab3b90, 0x88830b88,
    0x46ca8c46, 0xee29c7ee, 0xb8d36bb8, 0x143c2814,
    0xde79a7de, 0x5ee2bc5e, 0x0b1d160b, 0xdb76addb,
    0xe03bdbe0, 0x32566432, 0x3a4e743a, 0x0a1e140a,
    0x49db9249, 0x060a0c06, 0x246c4824, 0x5ce4b85c,
    0xc25d9fc2, 0xd36ebdd3, 0xacef43ac, 0x62a6c462,
    0x91a83991, 0x95a43195, 0xe437d3e4, 0x798bf279,
    0xe732d5e7, 0xc8438bc8, 0x37596e37, 0x6db7da6d,
    0x8d8c018d, 0xd564b1d5, 0x4ed29c4e, 0xa9e049a9,
    0x6cb4d86c, 0x56faac56, 0xf407f3f4, 0xea25cfea,
    0x65afca65, 0x7a8ef47a, 0xaee947ae, 0x08181008,
    0xbad56fba, 0x7888f078, 0x256f4a25, 0x2e725c2e,
    0x1c24381c, 0xa6f157a6, 0xb4c773b4, 0xc65197c6,
    0xe823cbe8, 0xdd7ca1dd, 0x749ce874, 0x1f213e1f,
    0x4bdd964b, 0xbddc61bd, 0x8b860d8b, 0x8a850f8a,
    0x7090e070, 0x3e427c3e, 0xb5c471b5, 0x66aacc66,
    0x48d89048, 0x03050603, 0xf601f7f6, 0x0e121c0e,
    0x61a3c261, 0x355f6a35, 0x57f9ae57, 0xb9d069b9,
    0x86911786, 0xc15899c1, 0x1d273a1d, 0x9eb9279e,
    0xe138d9e1, 0xf813ebf8, 0x98b32b98, 0x11332211,
    0x69bbd269, 0xd970a9d9, 0x8e89078e, 0x94a73394,
    0x9bb62d9b, 0x1e223c1e, 0x87921587, 0xe920c9e9,
    0xce4987ce, 0x55ffaa55, 0x28785028, 0xdf7aa5df,
    0x8c8f038c, 0xa1f859a1, 0x89800989, 0x0d171a0d,
    0xbfda65bf, 0xe631d7e6, 0x42c68442, 0x68b8d068,
    0x41c38241, 0x99b02999, 0x2d775a2d, 0x0f111e0f,
    0xb0cb7bb0, 0x54fca854, 0xbbd66dbb, 0x163a2c16], dtype=np.uint32)

AES_TE3 = np.array([
    0x6363a5c6, 0x7c7c84f8, 0x777799ee, 0x7b7b8df6,
    0xf2f20dff, 0x6b6bbdd6, 0x6f6fb1de, 0xc5c55491,
    0x30305060, 0x01010302, 0x6767a9ce, 0x2b2b7d56,
    0xfefe19e7, 0xd7d762b5, 0xababe64d, 0x76769aec,
    0xcaca458f, 0x82829d1f, 0xc9c94089, 0x7d7d87fa,
    0xfafa15ef, 0x5959ebb2, 0x4747c98e, 0xf0f00bfb,
    0xadadec41, 0xd4d467b3, 0xa2a2fd5f, 0xafafea45,
    0x9c9cbf23, 0xa4a4f753, 0x727296e4, 0xc0c05b9b,
    0xb7b7c275, 0xfdfd1ce1, 0x9393ae3d, 0x26266a4c,
    0x36365a6c, 0x3f3f417e, 0xf7f702f5, 0xcccc4f83,
    0x34345c68, 0xa5a5f451, 0xe5e534d1, 0xf1f108f9,
    0x717193e2, 0xd8d873ab, 0x31315362, 0x15153f2a,
    0x04040c08, 0xc7c75295, 0x23236546, 0xc3c35e9d,
    0x18182830, 0x9696a137, 0x05050f0a, 0x9a9ab52f,
    0x0707090e, 0x12123624, 0x80809b1b, 0xe2e23ddf,
    0xebeb26cd, 0x2727694e, 0xb2b2cd7f, 0x75759fea,
    0x09091b12, 0x83839e1d, 0x2c2c7458, 0x1a1a2e34,
    0x1b1b2d36, 0x6e6eb2dc, 0x5a5aeeb4, 0xa0a0fb5b,
    0x5252f6a4, 0x3b3b4d76, 0xd6d661b7, 0xb3b3ce7d,
    0x29297b52, 0xe3e33edd, 0x2f2f715e, 0x84849713,
    0x5353f5a6, 0xd1d168b9, 0x00000000, 0xeded2cc1,
    0x20206040, 0xfcfc1fe3, 0xb1b1c879, 0x5b5bedb6,
    0x6a6abed4, 0xcbcb468d, 0xbebed967, 0x39394b72,
    0x4a4ade94, 0x4c4cd498, 0x5858e8b0, 0xcfcf4a85,
    0xd0d06bbb, 0xefef2ac5, 0xaaaae54f, 0xfbfb16ed,
    0x4343c586, 0x4d4dd79a, 0x33335566, 0x85859411,
    0x4545cf8a, 0xf9f910e9, 0x02020604, 0x7f7f81fe,
    0x5050f0a0, 0x3c3c4478, 0x9f9fba25, 0xa8a8e34b,
    0x5151f3a2, 0xa3a3fe5d, 0x4040c080, 0x8f8f8a05,
    0x9292ad3f, 0x9d9dbc21, 0x38384870, 0xf5f504f1,
    0xbcbcdf63, 0xb6b6c177, 0xdada75af, 0x21216342,
    0x10103020, 0xffff1ae5, 0xf3f30efd, 0xd2d26dbf,
    0xcdcd4c81, 0x0c0c1418, 0x13133526, 0xecec2fc3,
    0x5f5fe1be, 0x9797a235, 0x4444cc88, 0x1717392e,
    0xc4c45793, 0xa7a7f255, 0x7e7e82fc, 0x3d3d477a,
    0x6464acc8, 0x5d5de7ba, 0x19192b32, 0x737395e6,
    0x6060a0c0, 0x81819819, 0x4f4fd19e, 0xdcdc7fa3,
    0x22226644, 0x2a2a7e54, 0x9090ab3b, 0x8888830b,
    0x4646ca8c, 0xeeee29c7, 0xb8b8d36b, 0x14143c28,
    0xdede79a7, 0x5e5ee2bc, 0x0b0b1d16, 0xdbdb76ad,
    0xe0e03bdb, 0x32325664, 0x3a3a4e74, 0x0a0a1e14,
    0x4949db92, 0x06060a0c, 0x24246c48, 0x5c5ce4b8,
    0xc2c25d9f, 0xd3d36ebd, 0xacacef43, 0x6262a6c4,
    0x9191a839, 0x9595a431, 0xe4e437d3, 0x79798bf2,
    0xe7e732d5, 0xc8c8438b, 0x3737596e, 0x6d6db7da,
    0x8d8d8c01, 0xd5d564b1, 0x4e4ed29c, 0xa9a9e049,
    0x6c6cb4d8, 0x5656faac, 0xf4f407f3, 0xeaea25cf,
    0x6565afca, 0x7a7a8ef4, 0xaeaee947, 0x08081810,
    0xbabad56f, 0x787888f0, 0x25256f4a, 0x2e2e725c,
    0x1c1c2438, 0xa6a6f157, 0xb4b4c773, 0xc6c65197,
    0xe8e823cb, 0xdddd7ca1, 0x74749ce8, 0x1f1f213e,
    0x4b4bdd96, 0xbdbddc61, 0x8b8b860d, 0x8a8a850f,
    0x707090e0, 0x3e3e427c, 0xb5b5c471, 0x6666aacc,
    0x4848d890, 0x03030506, 0xf6f601f7, 0x0e0e121c,
    0x6161a3c2, 0x35355f6a, 0x5757f9ae, 0xb9b9d069,
    0x86869117, 0xc1c15899, 0x1d1d273a, 0x9e9eb927,
    0xe1e138d9, 0xf8f813eb, 0x9898b32b, 0x11113322,
    0x6969bbd2, 0xd9d970a9, 0x8e8e8907, 0x9494a733,
    0x9b9bb62d, 0x1e1e223c, 0x87879215, 0xe9e920c9,
    0xcece4987, 0x5555ffaa, 0x28287850, 0xdfdf7aa5,
    0x8c8c8f03, 0xa1a1f859, 0x89898009, 0x0d0d171a,
    0xbfbfda65, 0xe6e631d7, 0x4242c684, 0x6868b8d0,
    0x4141c382, 0x9999b029, 0x2d2d775a, 0x0f0f111e,
    0xb0b0cb7b, 0x5454fca8, 0xbbbbd66d, 0x16163a2c], dtype=np.uint32)

AES_TE = np.array([AES_TE0, AES_TE1, AES_TE2, AES_TE3])

RCON = np.array([[0x00, 0x00, 0x00, 0x00], [0x01, 0x00, 0x00, 0x00],
                 [0x02, 0x00, 0x00, 0x00], [0x04, 0x00, 0x00, 0x00],
                 [0x08, 0x00, 0x00, 0x00], [0x10, 0x00, 0x00, 0x00],
                 [0x20, 0x00, 0x00, 0x00], [0x40, 0x00, 0x00, 0x00],
                 [0x80, 0x00, 0x00, 0x00], [0x1B, 0x00, 0x00, 0x00],
                 [0x36, 0x00, 0x00, 0x00]], dtype=np.uint8)

GF_MULT2 = np.array([
    0x00, 0x02, 0x04, 0x06, 0x08, 0x0a, 0x0c, 0x0e, 0x10, 0x12, 0x14, 0x16,
    0x18, 0x1a, 0x1c, 0x1e, 0x20, 0x22, 0x24, 0x26, 0x28, 0x2a, 0x2c, 0x2e,
    0x30, 0x32, 0x34, 0x36, 0x38, 0x3a, 0x3c, 0x3e, 0x40, 0x42, 0x44, 0x46,
    0x48, 0x4a, 0x4c, 0x4e, 0x50, 0x52, 0x54, 0x56, 0x58, 0x5a, 0x5c, 0x5e,
    0x60, 0x62, 0x64, 0x66, 0x68, 0x6a, 0x6c, 0x6e, 0x70, 0x72, 0x74, 0x76,
    0x78, 0x7a, 0x7c, 0x7e, 0x80, 0x82, 0x84, 0x86, 0x88, 0x8a, 0x8c, 0x8e,
    0x90, 0x92, 0x94, 0x96, 0x98, 0x9a, 0x9c, 0x9e, 0xa0, 0xa2, 0xa4, 0xa6,
    0xa8, 0xaa, 0xac, 0xae, 0xb0, 0xb2, 0xb4, 0xb6, 0xb8, 0xba, 0xbc, 0xbe,
    0xc0, 0xc2, 0xc4, 0xc6, 0xc8, 0xca, 0xcc, 0xce, 0xd0, 0xd2, 0xd4, 0xd6,
    0xd8, 0xda, 0xdc, 0xde, 0xe0, 0xe2, 0xe4, 0xe6, 0xe8, 0xea, 0xec, 0xee,
    0xf0, 0xf2, 0xf4, 0xf6, 0xf8, 0xfa, 0xfc, 0xfe, 0x1b, 0x19, 0x1f, 0x1d,
    0x13, 0x11, 0x17, 0x15, 0x0b, 0x09, 0x0f, 0x0d, 0x03, 0x01, 0x07, 0x05,
    0x3b, 0x39, 0x3f, 0x3d, 0x33, 0x31, 0x37, 0x35, 0x2b, 0x29, 0x2f, 0x2d,
    0x23, 0x21, 0x27, 0x25, 0x5b, 0x59, 0x5f, 0x5d, 0x53, 0x51, 0x57, 0x55,
    0x4b, 0x49, 0x4f, 0x4d, 0x43, 0x41, 0x47, 0x45, 0x7b, 0x79, 0x7f, 0x7d,
    0x73, 0x71, 0x77, 0x75, 0x6b, 0x69, 0x6f, 0x6d, 0x63, 0x61, 0x67, 0x65,
    0x9b, 0x99, 0x9f, 0x9d, 0x93, 0x91, 0x97, 0x95, 0x8b, 0x89, 0x8f, 0x8d,
    0x83, 0x81, 0x87, 0x85, 0xbb, 0xb9, 0xbf, 0xbd, 0xb3, 0xb1, 0xb7, 0xb5,
    0xab, 0xa9, 0xaf, 0xad, 0xa3, 0xa1, 0xa7, 0xa5, 0xdb, 0xd9, 0xdf, 0xdd,
    0xd3, 0xd1, 0xd7, 0xd5, 0xcb, 0xc9, 0xcf, 0xcd, 0xc3, 0xc1, 0xc7, 0xc5,
    0xfb, 0xf9, 0xff, 0xfd, 0xf3, 0xf1, 0xf7, 0xf5, 0xeb, 0xe9, 0xef, 0xed,
    0xe3, 0xe1, 0xe7, 0xe5
], dtype=np.uint8)

GF_INV_MULT2 = np.array([
    0x00, 0x8d, 0x01, 0x8c, 0x02, 0x8f, 0x03, 0x8e, 0x04, 0x89, 0x05,
    0x88, 0x06, 0x8b, 0x07, 0x8a, 0x08, 0x85, 0x09, 0x84, 0x0a, 0x87,
    0x0b, 0x86, 0x0c, 0x81, 0x0d, 0x80, 0x0e, 0x83, 0x0f, 0x82, 0x10,
    0x9d, 0x11, 0x9c, 0x12, 0x9f, 0x13, 0x9e, 0x14, 0x99, 0x15, 0x98,
    0x16, 0x9b, 0x17, 0x9a, 0x18, 0x95, 0x19, 0x94, 0x1a, 0x97, 0x1b,
    0x96, 0x1c, 0x91, 0x1d, 0x90, 0x1e, 0x93, 0x1f, 0x92, 0x20, 0xad,
    0x21, 0xac, 0x22, 0xaf, 0x23, 0xae, 0x24, 0xa9, 0x25, 0xa8, 0x26,
    0xab, 0x27, 0xaa, 0x28, 0xa5, 0x29, 0xa4, 0x2a, 0xa7, 0x2b, 0xa6,
    0x2c, 0xa1, 0x2d, 0xa0, 0x2e, 0xa3, 0x2f, 0xa2, 0x30, 0xbd, 0x31,
    0xbc, 0x32, 0xbf, 0x33, 0xbe, 0x34, 0xb9, 0x35, 0xb8, 0x36, 0xbb,
    0x37, 0xba, 0x38, 0xb5, 0x39, 0xb4, 0x3a, 0xb7, 0x3b, 0xb6, 0x3c,
    0xb1, 0x3d, 0xb0, 0x3e, 0xb3, 0x3f, 0xb2, 0x40, 0xcd, 0x41, 0xcc,
    0x42, 0xcf, 0x43, 0xce, 0x44, 0xc9, 0x45, 0xc8, 0x46, 0xcb, 0x47,
    0xca, 0x48, 0xc5, 0x49, 0xc4, 0x4a, 0xc7, 0x4b, 0xc6, 0x4c, 0xc1,
    0x4d, 0xc0, 0x4e, 0xc3, 0x4f, 0xc2, 0x50, 0xdd, 0x51, 0xdc, 0x52,
    0xdf, 0x53, 0xde, 0x54, 0xd9, 0x55, 0xd8, 0x56, 0xdb, 0x57, 0xda,
    0x58, 0xd5, 0x59, 0xd4, 0x5a, 0xd7, 0x5b, 0xd6, 0x5c, 0xd1, 0x5d,
    0xd0, 0x5e, 0xd3, 0x5f, 0xd2, 0x60, 0xed, 0x61, 0xec, 0x62, 0xef,
    0x63, 0xee, 0x64, 0xe9, 0x65, 0xe8, 0x66, 0xeb, 0x67, 0xea, 0x68,
    0xe5, 0x69, 0xe4, 0x6a, 0xe7, 0x6b, 0xe6, 0x6c, 0xe1, 0x6d, 0xe0,
    0x6e, 0xe3, 0x6f, 0xe2, 0x70, 0xfd, 0x71, 0xfc, 0x72, 0xff, 0x73,
    0xfe, 0x74, 0xf9, 0x75, 0xf8, 0x76, 0xfb, 0x77, 0xfa, 0x78, 0xf5,
    0x79, 0xf4, 0x7a, 0xf7, 0x7b, 0xf6, 0x7c, 0xf1, 0x7d, 0xf0, 0x7e,
    0xf3, 0x7f, 0xf2
], dtype=np.uint8)

GF_MULT3 = np.array([
    0x00, 0x03, 0x06, 0x05, 0x0c, 0x0f, 0x0a, 0x09, 0x18, 0x1b, 0x1e, 0x1d,
    0x14, 0x17, 0x12, 0x11, 0x30, 0x33, 0x36, 0x35, 0x3c, 0x3f, 0x3a, 0x39,
    0x28, 0x2b, 0x2e, 0x2d, 0x24, 0x27, 0x22, 0x21, 0x60, 0x63, 0x66, 0x65,
    0x6c, 0x6f, 0x6a, 0x69, 0x78, 0x7b, 0x7e, 0x7d, 0x74, 0x77, 0x72, 0x71,
    0x50, 0x53, 0x56, 0x55, 0x5c, 0x5f, 0x5a, 0x59, 0x48, 0x4b, 0x4e, 0x4d,
    0x44, 0x47, 0x42, 0x41, 0xc0, 0xc3, 0xc6, 0xc5, 0xcc, 0xcf, 0xca, 0xc9,
    0xd8, 0xdb, 0xde, 0xdd, 0xd4, 0xd7, 0xd2, 0xd1, 0xf0, 0xf3, 0xf6, 0xf5,
    0xfc, 0xff, 0xfa, 0xf9, 0xe8, 0xeb, 0xee, 0xed, 0xe4, 0xe7, 0xe2, 0xe1,
    0xa0, 0xa3, 0xa6, 0xa5, 0xac, 0xaf, 0xaa, 0xa9, 0xb8, 0xbb, 0xbe, 0xbd,
    0xb4, 0xb7, 0xb2, 0xb1, 0x90, 0x93, 0x96, 0x95, 0x9c, 0x9f, 0x9a, 0x99,
    0x88, 0x8b, 0x8e, 0x8d, 0x84, 0x87, 0x82, 0x81, 0x9b, 0x98, 0x9d, 0x9e,
    0x97, 0x94, 0x91, 0x92, 0x83, 0x80, 0x85, 0x86, 0x8f, 0x8c, 0x89, 0x8a,
    0xab, 0xa8, 0xad, 0xae, 0xa7, 0xa4, 0xa1, 0xa2, 0xb3, 0xb0, 0xb5, 0xb6,
    0xbf, 0xbc, 0xb9, 0xba, 0xfb, 0xf8, 0xfd, 0xfe, 0xf7, 0xf4, 0xf1, 0xf2,
    0xe3, 0xe0, 0xe5, 0xe6, 0xef, 0xec, 0xe9, 0xea, 0xcb, 0xc8, 0xcd, 0xce,
    0xc7, 0xc4, 0xc1, 0xc2, 0xd3, 0xd0, 0xd5, 0xd6, 0xdf, 0xdc, 0xd9, 0xda,
    0x5b, 0x58, 0x5d, 0x5e, 0x57, 0x54, 0x51, 0x52, 0x43, 0x40, 0x45, 0x46,
    0x4f, 0x4c, 0x49, 0x4a, 0x6b, 0x68, 0x6d, 0x6e, 0x67, 0x64, 0x61, 0x62,
    0x73, 0x70, 0x75, 0x76, 0x7f, 0x7c, 0x79, 0x7a, 0x3b, 0x38, 0x3d, 0x3e,
    0x37, 0x34, 0x31, 0x32, 0x23, 0x20, 0x25, 0x26, 0x2f, 0x2c, 0x29, 0x2a,
    0x0b, 0x08, 0x0d, 0x0e, 0x07, 0x04, 0x01, 0x02, 0x13, 0x10, 0x15, 0x16,
    0x1f, 0x1c, 0x19, 0x1a
], dtype=np.uint8)

GF_INV_MULT3 = np.array([
    0x00, 0xf6, 0xf7, 0x01, 0xf5, 0x03, 0x02, 0xf4, 0xf1, 0x07, 0x06,
    0xf0, 0x04, 0xf2, 0xf3, 0x05, 0xf9, 0x0f, 0x0e, 0xf8, 0x0c, 0xfa,
    0xfb, 0x0d, 0x08, 0xfe, 0xff, 0x09, 0xfd, 0x0b, 0x0a, 0xfc, 0xe9,
    0x1f, 0x1e, 0xe8, 0x1c, 0xea, 0xeb, 0x1d, 0x18, 0xee, 0xef, 0x19,
    0xed, 0x1b, 0x1a, 0xec, 0x10, 0xe6, 0xe7, 0x11, 0xe5, 0x13, 0x12,
    0xe4, 0xe1, 0x17, 0x16, 0xe0, 0x14, 0xe2, 0xe3, 0x15, 0xc9, 0x3f,
    0x3e, 0xc8, 0x3c, 0xca, 0xcb, 0x3d, 0x38, 0xce, 0xcf, 0x39, 0xcd,
    0x3b, 0x3a, 0xcc, 0x30, 0xc6, 0xc7, 0x31, 0xc5, 0x33, 0x32, 0xc4,
    0xc1, 0x37, 0x36, 0xc0, 0x34, 0xc2, 0xc3, 0x35, 0x20, 0xd6, 0xd7,
    0x21, 0xd5, 0x23, 0x22, 0xd4, 0xd1, 0x27, 0x26, 0xd0, 0x24, 0xd2,
    0xd3, 0x25, 0xd9, 0x2f, 0x2e, 0xd8, 0x2c, 0xda, 0xdb, 0x2d, 0x28,
    0xde, 0xdf, 0x29, 0xdd, 0x2b, 0x2a, 0xdc, 0x89, 0x7f, 0x7e, 0x88,
    0x7c, 0x8a, 0x8b, 0x7d, 0x78, 0x8e, 0x8f, 0x79, 0x8d, 0x7b, 0x7a,
    0x8c, 0x70, 0x86, 0x87, 0x71, 0x85, 0x73, 0x72, 0x84, 0x81, 0x77,
    0x76, 0x80, 0x74, 0x82, 0x83, 0x75, 0x60, 0x96, 0x97, 0x61, 0x95,
    0x63, 0x62, 0x94, 0x91, 0x67, 0x66, 0x90, 0x64, 0x92, 0x93, 0x65,
    0x99, 0x6f, 0x6e, 0x98, 0x6c, 0x9a, 0x9b, 0x6d, 0x68, 0x9e, 0x9f,
    0x69, 0x9d, 0x6b, 0x6a, 0x9c, 0x40, 0xb6, 0xb7, 0x41, 0xb5, 0x43,
    0x42, 0xb4, 0xb1, 0x47, 0x46, 0xb0, 0x44, 0xb2, 0xb3, 0x45, 0xb9,
    0x4f, 0x4e, 0xb8, 0x4c, 0xba, 0xbb, 0x4d, 0x48, 0xbe, 0xbf, 0x49,
    0xbd, 0x4b, 0x4a, 0xbc, 0xa9, 0x5f, 0x5e, 0xa8, 0x5c, 0xaa, 0xab,
    0x5d, 0x58, 0xae, 0xaf, 0x59, 0xad, 0x5b, 0x5a, 0xac, 0x50, 0xa6,
    0xa7, 0x51, 0xa5, 0x53, 0x52, 0xa4, 0xa1, 0x57, 0x56, 0xa0, 0x54,
    0xa2, 0xa3, 0x55
], dtype=np.uint8)

GF_MULT9 = np.array([
    0x00, 0x09, 0x12, 0x1b, 0x24, 0x2d, 0x36, 0x3f, 0x48, 0x41, 0x5a, 0x53,
    0x6c, 0x65, 0x7e, 0x77, 0x90, 0x99, 0x82, 0x8b, 0xb4, 0xbd, 0xa6, 0xaf,
    0xd8, 0xd1, 0xca, 0xc3, 0xfc, 0xf5, 0xee, 0xe7, 0x3b, 0x32, 0x29, 0x20,
    0x1f, 0x16, 0x0d, 0x04, 0x73, 0x7a, 0x61, 0x68, 0x57, 0x5e, 0x45, 0x4c,
    0xab, 0xa2, 0xb9, 0xb0, 0x8f, 0x86, 0x9d, 0x94, 0xe3, 0xea, 0xf1, 0xf8,
    0xc7, 0xce, 0xd5, 0xdc, 0x76, 0x7f, 0x64, 0x6d, 0x52, 0x5b, 0x40, 0x49,
    0x3e, 0x37, 0x2c, 0x25, 0x1a, 0x13, 0x08, 0x01, 0xe6, 0xef, 0xf4, 0xfd,
    0xc2, 0xcb, 0xd0, 0xd9, 0xae, 0xa7, 0xbc, 0xb5, 0x8a, 0x83, 0x98, 0x91,
    0x4d, 0x44, 0x5f, 0x56, 0x69, 0x60, 0x7b, 0x72, 0x05, 0x0c, 0x17, 0x1e,
    0x21, 0x28, 0x33, 0x3a, 0xdd, 0xd4, 0xcf, 0xc6, 0xf9, 0xf0, 0xeb, 0xe2,
    0x95, 0x9c, 0x87, 0x8e, 0xb1, 0xb8, 0xa3, 0xaa, 0xec, 0xe5, 0xfe, 0xf7,
    0xc8, 0xc1, 0xda, 0xd3, 0xa4, 0xad, 0xb6, 0xbf, 0x80, 0x89, 0x92, 0x9b,
    0x7c, 0x75, 0x6e, 0x67, 0x58, 0x51, 0x4a, 0x43, 0x34, 0x3d, 0x26, 0x2f,
    0x10, 0x19, 0x02, 0x0b, 0xd7, 0xde, 0xc5, 0xcc, 0xf3, 0xfa, 0xe1, 0xe8,
    0x9f, 0x96, 0x8d, 0x84, 0xbb, 0xb2, 0xa9, 0xa0, 0x47, 0x4e, 0x55, 0x5c,
    0x63, 0x6a, 0x71, 0x78, 0x0f, 0x06, 0x1d, 0x14, 0x2b, 0x22, 0x39, 0x30,
    0x9a, 0x93, 0x88, 0x81, 0xbe, 0xb7, 0xac, 0xa5, 0xd2, 0xdb, 0xc0, 0xc9,
    0xf6, 0xff, 0xe4, 0xed, 0x0a, 0x03, 0x18, 0x11, 0x2e, 0x27, 0x3c, 0x35,
    0x42, 0x4b, 0x50, 0x59, 0x66, 0x6f, 0x74, 0x7d, 0xa1, 0xa8, 0xb3, 0xba,
    0x85, 0x8c, 0x97, 0x9e, 0xe9, 0xe0, 0xfb, 0xf2, 0xcd, 0xc4, 0xdf, 0xd6,
    0x31, 0x38, 0x23, 0x2a, 0x15, 0x1c, 0x07, 0x0e, 0x79, 0x70, 0x6b, 0x62,
    0x5d, 0x54, 0x4f, 0x46
], dtype=np.uint8)

GF_MULT11 = np.array([
    0x00, 0x0b, 0x16, 0x1d, 0x2c, 0x27, 0x3a, 0x31, 0x58, 0x53, 0x4e, 0x45,
    0x74, 0x7f, 0x62, 0x69, 0xb0, 0xbb, 0xa6, 0xad, 0x9c, 0x97, 0x8a, 0x81,
    0xe8, 0xe3, 0xfe, 0xf5, 0xc4, 0xcf, 0xd2, 0xd9, 0x7b, 0x70, 0x6d, 0x66,
    0x57, 0x5c, 0x41, 0x4a, 0x23, 0x28, 0x35, 0x3e, 0x0f, 0x04, 0x19, 0x12,
    0xcb, 0xc0, 0xdd, 0xd6, 0xe7, 0xec, 0xf1, 0xfa, 0x93, 0x98, 0x85, 0x8e,
    0xbf, 0xb4, 0xa9, 0xa2, 0xf6, 0xfd, 0xe0, 0xeb, 0xda, 0xd1, 0xcc, 0xc7,
    0xae, 0xa5, 0xb8, 0xb3, 0x82, 0x89, 0x94, 0x9f, 0x46, 0x4d, 0x50, 0x5b,
    0x6a, 0x61, 0x7c, 0x77, 0x1e, 0x15, 0x08, 0x03, 0x32, 0x39, 0x24, 0x2f,
    0x8d, 0x86, 0x9b, 0x90, 0xa1, 0xaa, 0xb7, 0xbc, 0xd5, 0xde, 0xc3, 0xc8,
    0xf9, 0xf2, 0xef, 0xe4, 0x3d, 0x36, 0x2b, 0x20, 0x11, 0x1a, 0x07, 0x0c,
    0x65, 0x6e, 0x73, 0x78, 0x49, 0x42, 0x5f, 0x54, 0xf7, 0xfc, 0xe1, 0xea,
    0xdb, 0xd0, 0xcd, 0xc6, 0xaf, 0xa4, 0xb9, 0xb2, 0x83, 0x88, 0x95, 0x9e,
    0x47, 0x4c, 0x51, 0x5a, 0x6b, 0x60, 0x7d, 0x76, 0x1f, 0x14, 0x09, 0x02,
    0x33, 0x38, 0x25, 0x2e, 0x8c, 0x87, 0x9a, 0x91, 0xa0, 0xab, 0xb6, 0xbd,
    0xd4, 0xdf, 0xc2, 0xc9, 0xf8, 0xf3, 0xee, 0xe5, 0x3c, 0x37, 0x2a, 0x21,
    0x10, 0x1b, 0x06, 0x0d, 0x64, 0x6f, 0x72, 0x79, 0x48, 0x43, 0x5e, 0x55,
    0x01, 0x0a, 0x17, 0x1c, 0x2d, 0x26, 0x3b, 0x30, 0x59, 0x52, 0x4f, 0x44,
    0x75, 0x7e, 0x63, 0x68, 0xb1, 0xba, 0xa7, 0xac, 0x9d, 0x96, 0x8b, 0x80,
    0xe9, 0xe2, 0xff, 0xf4, 0xc5, 0xce, 0xd3, 0xd8, 0x7a, 0x71, 0x6c, 0x67,
    0x56, 0x5d, 0x40, 0x4b, 0x22, 0x29, 0x34, 0x3f, 0x0e, 0x05, 0x18, 0x13,
    0xca, 0xc1, 0xdc, 0xd7, 0xe6, 0xed, 0xf0, 0xfb, 0x92, 0x99, 0x84, 0x8f,
    0xbe, 0xb5, 0xa8, 0xa3
], dtype=np.uint8)

GF_MULT13 = np.array([
    0x00, 0x0d, 0x1a, 0x17, 0x34, 0x39, 0x2e, 0x23, 0x68, 0x65, 0x72, 0x7f,
    0x5c, 0x51, 0x46, 0x4b, 0xd0, 0xdd, 0xca, 0xc7, 0xe4, 0xe9, 0xfe, 0xf3,
    0xb8, 0xb5, 0xa2, 0xaf, 0x8c, 0x81, 0x96, 0x9b, 0xbb, 0xb6, 0xa1, 0xac,
    0x8f, 0x82, 0x95, 0x98, 0xd3, 0xde, 0xc9, 0xc4, 0xe7, 0xea, 0xfd, 0xf0,
    0x6b, 0x66, 0x71, 0x7c, 0x5f, 0x52, 0x45, 0x48, 0x03, 0x0e, 0x19, 0x14,
    0x37, 0x3a, 0x2d, 0x20, 0x6d, 0x60, 0x77, 0x7a, 0x59, 0x54, 0x43, 0x4e,
    0x05, 0x08, 0x1f, 0x12, 0x31, 0x3c, 0x2b, 0x26, 0xbd, 0xb0, 0xa7, 0xaa,
    0x89, 0x84, 0x93, 0x9e, 0xd5, 0xd8, 0xcf, 0xc2, 0xe1, 0xec, 0xfb, 0xf6,
    0xd6, 0xdb, 0xcc, 0xc1, 0xe2, 0xef, 0xf8, 0xf5, 0xbe, 0xb3, 0xa4, 0xa9,
    0x8a, 0x87, 0x90, 0x9d, 0x06, 0x0b, 0x1c, 0x11, 0x32, 0x3f, 0x28, 0x25,
    0x6e, 0x63, 0x74, 0x79, 0x5a, 0x57, 0x40, 0x4d, 0xda, 0xd7, 0xc0, 0xcd,
    0xee, 0xe3, 0xf4, 0xf9, 0xb2, 0xbf, 0xa8, 0xa5, 0x86, 0x8b, 0x9c, 0x91,
    0x0a, 0x07, 0x10, 0x1d, 0x3e, 0x33, 0x24, 0x29, 0x62, 0x6f, 0x78, 0x75,
    0x56, 0x5b, 0x4c, 0x41, 0x61, 0x6c, 0x7b, 0x76, 0x55, 0x58, 0x4f, 0x42,
    0x09, 0x04, 0x13, 0x1e, 0x3d, 0x30, 0x27, 0x2a, 0xb1, 0xbc, 0xab, 0xa6,
    0x85, 0x88, 0x9f, 0x92, 0xd9, 0xd4, 0xc3, 0xce, 0xed, 0xe0, 0xf7, 0xfa,
    0xb7, 0xba, 0xad, 0xa0, 0x83, 0x8e, 0x99, 0x94, 0xdf, 0xd2, 0xc5, 0xc8,
    0xeb, 0xe6, 0xf1, 0xfc, 0x67, 0x6a, 0x7d, 0x70, 0x53, 0x5e, 0x49, 0x44,
    0x0f, 0x02, 0x15, 0x18, 0x3b, 0x36, 0x21, 0x2c, 0x0c, 0x01, 0x16, 0x1b,
    0x38, 0x35, 0x22, 0x2f, 0x64, 0x69, 0x7e, 0x73, 0x50, 0x5d, 0x4a, 0x47,
    0xdc, 0xd1, 0xc6, 0xcb, 0xe8, 0xe5, 0xf2, 0xff, 0xb4, 0xb9, 0xae, 0xa3,
    0x80, 0x8d, 0x9a, 0x97
], dtype=np.uint8)

GF_MULT14 = np.array([
    0x00, 0x0e, 0x1c, 0x12, 0x38, 0x36, 0x24, 0x2a, 0x70, 0x7e, 0x6c, 0x62,
    0x48, 0x46, 0x54, 0x5a, 0xe0, 0xee, 0xfc, 0xf2, 0xd8, 0xd6, 0xc4, 0xca,
    0x90, 0x9e, 0x8c, 0x82, 0xa8, 0xa6, 0xb4, 0xba, 0xdb, 0xd5, 0xc7, 0xc9,
    0xe3, 0xed, 0xff, 0xf1, 0xab, 0xa5, 0xb7, 0xb9, 0x93, 0x9d, 0x8f, 0x81,
    0x3b, 0x35, 0x27, 0x29, 0x03, 0x0d, 0x1f, 0x11, 0x4b, 0x45, 0x57, 0x59,
    0x73, 0x7d, 0x6f, 0x61, 0xad, 0xa3, 0xb1, 0xbf, 0x95, 0x9b, 0x89, 0x87,
    0xdd, 0xd3, 0xc1, 0xcf, 0xe5, 0xeb, 0xf9, 0xf7, 0x4d, 0x43, 0x51, 0x5f,
    0x75, 0x7b, 0x69, 0x67, 0x3d, 0x33, 0x21, 0x2f, 0x05, 0x0b, 0x19, 0x17,
    0x76, 0x78, 0x6a, 0x64, 0x4e, 0x40, 0x52, 0x5c, 0x06, 0x08, 0x1a, 0x14,
    0x3e, 0x30, 0x22, 0x2c, 0x96, 0x98, 0x8a, 0x84, 0xae, 0xa0, 0xb2, 0xbc,
    0xe6, 0xe8, 0xfa, 0xf4, 0xde, 0xd0, 0xc2, 0xcc, 0x41, 0x4f, 0x5d, 0x53,
    0x79, 0x77, 0x65, 0x6b, 0x31, 0x3f, 0x2d, 0x23, 0x09, 0x07, 0x15, 0x1b,
    0xa1, 0xaf, 0xbd, 0xb3, 0x99, 0x97, 0x85, 0x8b, 0xd1, 0xdf, 0xcd, 0xc3,
    0xe9, 0xe7, 0xf5, 0xfb, 0x9a, 0x94, 0x86, 0x88, 0xa2, 0xac, 0xbe, 0xb0,
    0xea, 0xe4, 0xf6, 0xf8, 0xd2, 0xdc, 0xce, 0xc0, 0x7a, 0x74, 0x66, 0x68,
    0x42, 0x4c, 0x5e, 0x50, 0x0a, 0x04, 0x16, 0x18, 0x32, 0x3c, 0x2e, 0x20,
    0xec, 0xe2, 0xf0, 0xfe, 0xd4, 0xda, 0xc8, 0xc6, 0x9c, 0x92, 0x80, 0x8e,
    0xa4, 0xaa, 0xb8, 0xb6, 0x0c, 0x02, 0x10, 0x1e, 0x34, 0x3a, 0x28, 0x26,
    0x7c, 0x72, 0x60, 0x6e, 0x44, 0x4a, 0x58, 0x56, 0x37, 0x39, 0x2b, 0x25,
    0x0f, 0x01, 0x13, 0x1d, 0x47, 0x49, 0x5b, 0x55, 0x7f, 0x71, 0x63, 0x6d,
    0xd7, 0xd9, 0xcb, 0xc5, 0xef, 0xe1, 0xf3, 0xfd, 0xa7, 0xa9, 0xbb, 0xb5,
    0x9f, 0x91, 0x83, 0x8d
], dtype=np.uint8)


# fmt: on


def aes_nist_key() -> np.ndarray:
    """
    The example key used by the NIST description of AES.

    This key is so common in SCA that it deserves to be available in any SCA
    framework!

    .. versionadded:: 1.3.0

    """
    return np.array(
        [
            0x2B,
            0x7E,
            0x15,
            0x16,
            0x28,
            0xAE,
            0xD2,
            0xA6,
            0xAB,
            0xF7,
            0x15,
            0x88,
            0x09,
            0xCF,
            0x4F,
            0x3C,
        ],
        dtype=np.uint8,
    )


def aes_sbox(x: np.ndarray) -> np.ndarray:
    """Vectorized AES substitution operation model (aka., SBOX)

    :Examples:

    >>> aes_sbox(np.arange(10))
    array([ 99, 124, 119, 123, 242, 107, 111, 197,  48,   1], dtype=uint8)
    """
    return AES_SBOX_TABLE[x]


def aes_inv_sbox(x: np.ndarray) -> np.ndarray:
    """
    Vectorized AES inverse substitution (aka., SBOX^{-1})
    """
    return AES_INV_SBOX_TABLE[x]


def aes_sbox_leakage(k, p: np.ndarray, noise=0.1) -> np.ndarray:
    """Ideal leakage model of first Subbytes operation of the AES

    This function is a vectorized version of
    ``hamming_weight(aes_sbox(k ^ p))``, with some optional noise.

    :param noise: Noise magnitude (gaussian noise is added)
    """
    err = np.random.normal(0, noise, size=p.shape[0])
    return hamming_weight(aes_sbox(k ^ p)) + err


def aes_t_table(i: int, x: np.ndarray) -> np.ndarray:
    """Vectorized AES T-Table model

    >>> aes_t_table(0, np.arange(3))
    array([3328402341, 4168907908, 4000806809], dtype=uint32)
    """
    return AES_TE[i][x]


def aes_t_indices(i):
    """Return the T-table indice used by the x-th key byte"""
    return i & 3


def aes_expand_key(key: np.array) -> np.ndarray:
    """
    Compute AES intermediate round keys.

    :param key: The key to expand (a numpy array of 16 bytes).
    :return: A (11, 16) array representing the AES round keys.
    """
    round_keys = [key]
    for i in range(1, 11):
        round_keys.append(aes_expand_key_step(round_keys[-1], i - 1))
    return np.array(round_keys, dtype=np.uint8)


def aes_expand_key_step(round_key: np.array, aes_round: int) -> np.array:
    """
    Compute one iteration of the key expansion.
    """
    w = round_key.reshape(4, 4)
    rk = []
    rk.append(aes_sbox(np.roll(w[-1], -1)) ^ w[0] ^ RCON[aes_round + 1])
    for j in range(3):
        rk.append(rk[-1] ^ w[j + 1])
    return np.array(rk, dtype=np.uint8).reshape(16)


def aes_inv_expand_key(round_key: np.array) -> np.ndarray:
    """
    Compute AES intermediate round keys.

    :param key: The key to expand (a numpy array of 16 bytes).
    :return: A (11, 16) array representing the AES round keys. Index 10 is the
        key given as input and index 0 is the 0th round key. Thus, the array
        returned can be used for an AES encryption "as this".
    """
    round_keys = [round_key]
    for i in range(9, -1, -1):
        round_keys.append(aes_inv_expand_key_step(round_keys[-1], i + 1))
    return np.array(round_keys, dtype=np.uint8)[::-1]


def aes_inv_expand_key_step(round_key: np.array, aes_round: int) -> np.array:
    """Compute one iteration of the reverse key expansion."""
    w = round_key.reshape(4, 4)
    rk = []
    for j in range(3, 0, -1):
        rk.append(w[j] ^ w[j - 1])
    rk.append(w[0] ^ np.roll(aes_sbox(rk[0]), -1) ^ RCON[aes_round])
    return np.array(rk, dtype=np.uint8)[::-1].reshape(16)


class AesOps:
    """
    Set of basic operations on an AES state (i.e., a 4x4 byte matrix).

    The operations are all static methods, this class is used only for
    namespacing purposes.

    .. versionadded:: 1.3.0

    """

    @staticmethod
    def add_round_key(state: np.ndarray, rk: np.array) -> np.ndarray:
        return state ^ rk

    @staticmethod
    def sub_bytes(state: np.ndarray) -> np.ndarray:
        return aes_sbox(state)

    @staticmethod
    def inv_sub_bytes(state: np.ndarray) -> np.ndarray:
        return aes_inv_sbox(state)

    @staticmethod
    def shift_rows(states: np.ndarray) -> np.ndarray:
        states = states.reshape((states.shape[0], 4, 4))
        s0 = states[:, :, 0].reshape(states.shape[0], 4)
        s1 = states[:, :, 1].reshape(states.shape[0], 4)
        s2 = states[:, :, 2].reshape(states.shape[0], 4)
        s3 = states[:, :, 3].reshape(states.shape[0], 4)

        r1 = np.roll(s1, -1, axis=1)
        r2 = np.roll(s2, -2, axis=1)
        r3 = np.roll(s3, -3, axis=1)

        res = np.hstack((s0, r1, r2, r3)).reshape((states.shape[0], 4, 4))
        return res.transpose((0, 2, 1)).reshape((states.shape[0], 16))

    @staticmethod
    def inv_shift_rows(states: np.ndarray) -> np.ndarray:
        states = states.reshape(states.shape[0], 4, 4)
        s0 = states[:, :, 0].reshape(states.shape[0], 4)
        s1 = states[:, :, 1].reshape(states.shape[0], 4)
        s2 = states[:, :, 2].reshape(states.shape[0], 4)
        s3 = states[:, :, 3].reshape(states.shape[0], 4)

        r1 = np.roll(s1, 1, axis=1)
        r2 = np.roll(s2, 2, axis=1)
        r3 = np.roll(s3, 3, axis=1)

        res = np.hstack((s0, r1, r2, r3)).reshape(states.shape[0], 4, 4)
        return res.transpose((0, 2, 1)).reshape(states.shape[0], 16)

    @staticmethod
    def mix_columns(states: np.ndarray) -> np.ndarray:
        r = states.reshape((states.shape[0], 4, 4))
        res = np.zeros(r.shape, dtype=np.uint8)

        # NOTE: manual indentation for readability
        # @formatter:off
        # 2 * a0 + a3 + a2 + 3 * a1
        res[:, :, 0] = (
            GF_MULT2[r[:, :, 0]] ^ r[:, :, 3] ^ r[:, :, 2] ^ GF_MULT3[r[:, :, 1]]
        )
        # 2 * a1 + a0 + a3 + 3 * a2
        res[:, :, 1] = (
            GF_MULT2[r[:, :, 1]] ^ r[:, :, 0] ^ r[:, :, 3] ^ GF_MULT3[r[:, :, 2]]
        )
        # 2 * a2 + a1 + a0 + 3 * a3
        res[:, :, 2] = (
            GF_MULT2[r[:, :, 2]] ^ r[:, :, 1] ^ r[:, :, 0] ^ GF_MULT3[r[:, :, 3]]
        )
        # 2 * a3 + a2 + a1 + 3 * a0
        res[:, :, 3] = (
            GF_MULT2[r[:, :, 3]] ^ r[:, :, 2] ^ r[:, :, 1] ^ GF_MULT3[r[:, :, 0]]
        )
        # @formatter:on
        return res.reshape(states.shape[0], 16)

    @staticmethod
    def inv_mix_columns(states: np.ndarray) -> np.ndarray:
        r = states.reshape((states.shape[0], 4, 4))
        res = np.zeros(r.shape, dtype=np.uint8)

        # NOTE: manual indentation for readability
        # @formatter:off
        # 14 * a0 + 9 * a3 + 13 * a2 + 11 * a1
        res[:, :, 0] = (
            GF_MULT14[r[:, :, 0]]
            ^ GF_MULT9[r[:, :, 3]]
            ^ GF_MULT13[r[:, :, 2]]
            ^ GF_MULT11[r[:, :, 1]]
        )
        # 14 * a1 + 9 * a0 + 13 * a3 + 11 * a2
        res[:, :, 1] = (
            GF_MULT14[r[:, :, 1]]
            ^ GF_MULT9[r[:, :, 0]]
            ^ GF_MULT13[r[:, :, 3]]
            ^ GF_MULT11[r[:, :, 2]]
        )
        # 14 * a2 + 9 * a1 + 13 * a0 + 11 * a3
        res[:, :, 2] = (
            GF_MULT14[r[:, :, 2]]
            ^ GF_MULT9[r[:, :, 1]]
            ^ GF_MULT13[r[:, :, 0]]
            ^ GF_MULT11[r[:, :, 3]]
        )
        # 14 * a3 + 9 * a2 + 13 *a1 + 11 * a0
        res[:, :, 3] = (
            GF_MULT14[r[:, :, 3]]
            ^ GF_MULT9[r[:, :, 2]]
            ^ GF_MULT13[r[:, :, 1]]
            ^ GF_MULT11[r[:, :, 0]]
        )
        # @formatter:on
        return res.reshape(states.shape[0], 16)


class AES:
    """
    Flexible execution model of the AES cipher.

    .. versionadded:: 1.3.0

    """

    def __init__(self, key: np.array):
        """Initialize an AES cipher with a given key."""
        self.key = key
        self._round_keys = aes_expand_key(key)

    @classmethod
    def from_round_key(cls, key: np.ndarray, round: int):
        """Initialize an AES cipher from a specific round key."""
        assert 0 <= round <= 10, "invalid AES round given"
        # We finish the key expansion
        for i in range(round, 10):
            key = aes_expand_key_step(key, i)
        # Then, we revert it to generate the round keys
        round_keys = aes_inv_expand_key(key)
        return cls(round_keys[0])

    def encrypt(
        self,
        states: np.ndarray,
        start_after="start",
        stop_after=AesOps.add_round_key,
        start_round=0,
        stop_round=10,
    ):
        """
        Perform an AES encryption of the given initial states.

        Specific start and stop rounds can be passed to this function.

        If start_round is 0 and stop_round if 10, a normal encryption is
        performed.

        :param states: input states (a numpy array (N, 16) of np.uint8)
        :param start_round: Encrypt from this round.
        :param stop_round: Encrypt until this round.
        :param start_after: A method of
            :py:class:`secbench.processing.aes.AesOps` indicating the start
            point within the `start_round`, (default: 'start')
        :param stop_after: Last operation computed in the stop round
            (a method of :py:class:`secbench.processing.aes.AesOps`).

        :return: A (N, 16) array corresponding to the final
            (or intermediates) states.

        """
        operations = [
            "start",
            AesOps.sub_bytes,
            AesOps.shift_rows,
            AesOps.mix_columns,
            AesOps.add_round_key,
        ]
        if start_round > stop_round:
            raise ValueError("Invalid start and stop rounds")
        if stop_round > 10:
            raise ValueError("stop_round too large")

        if start_round == 0:
            states = AesOps.add_round_key(states, self._round_keys[0])
            if stop_round == 0:
                return states
        else:
            # Finish the starting round
            for i in range(operations.index(start_after) + 1, len(operations)):
                if operations[i] != AesOps.add_round_key:
                    states = operations[i](states)
                else:
                    states = operations[i](states, self._round_keys[start_round])
        for i in range(start_round + 1, stop_round):
            states = AesOps.sub_bytes(states)
            states = AesOps.shift_rows(states)
            states = AesOps.mix_columns(states)
            states = AesOps.add_round_key(states, self._round_keys[i])

        for i in range(1, operations.index(stop_after) + 1):
            if stop_round == 10 and i == operations.index(AesOps.mix_columns):
                continue
            if operations[i] != AesOps.add_round_key:
                states = operations[i](states)
            else:
                states = operations[i](states, self._round_keys[stop_round])
        return states

    def decrypt(
        self,
        states: np.ndarray,
        start_after="start",
        stop_after=AesOps.inv_sub_bytes,
        start_round=10,
        stop_round=0,
    ):
        """
        Perform an AES decryption of the given initial states.

        This method works exactly as
        :py:meth:`secbench.processing.aes.AES.encrypt`.
        Please refer to its documentation.
        """
        operations = [
            "start",
            AesOps.add_round_key,
            AesOps.inv_mix_columns,
            AesOps.inv_shift_rows,
            AesOps.inv_sub_bytes,
        ]
        if start_round < stop_round:
            raise ValueError(
                "start_round should be greater than \
                              stop_round"
            )
        if start_round > 10:
            raise ValueError("start_round too high")

        if stop_round < 0:
            raise ValueError("invalid stop_round")

        for i in range(operations.index(start_after) + 1, len(operations)):
            if start_round == 10 and i == operations.index(AesOps.inv_mix_columns):
                continue
            if operations[i] != AesOps.add_round_key:
                states = operations[i](states)
            else:
                states = operations[i](states, self._round_keys[start_round])
            # take into account stopping at round 10
            if stop_round == 10 and stop_after == operations[i]:
                return states

        for i in range(start_round - 1, stop_round, -1):
            states = AesOps.add_round_key(states, self._round_keys[i])
            states = AesOps.inv_mix_columns(states)
            states = AesOps.inv_shift_rows(states)
            states = AesOps.inv_sub_bytes(states)

        # Finish the starting round
        if stop_round == 0:
            states = AesOps.add_round_key(states, self._round_keys[stop_round])
        else:
            for i in range(1, operations.index(stop_after) + 1):
                if operations[i] != AesOps.add_round_key:
                    states = operations[i](states)
                else:
                    states = operations[i](states, self._round_keys[stop_round])
        return states


# ===
# Helpers for side-channel analysis (biased input generation).
# ===


def generate_plaintexts(
    n_traces: int,
    random_rows: Optional[List[int]] = None,
    random_cols: Optional[List[int]] = None,
    random_bytes: Optional[List[int]] = None,
    seed: Optional[int] = None,
    fixed_value: int = 0,
) -> np.ndarray:
    """
    Flexible generation of AES plaintexts.

    This function allows to easily select which part of the plaintext are randomized.

    It is important to understand that a flat plaintext is transformed in a
    column representation. Namely, the plaintext
    ``[P0 P1 P2 P3 P4 P5 P6 P7 P8 P9 P10 P11 P12 P13 P14 P15]`` is manipulated
    by the AES as:

    .. code-block:: text

             cols ->
        rows P0  P4  P8  P12
          |  P1  P5  P9  P13
          v  P2  P6  P10 P15
             P3  P7  P11 P16

    :param n_traces: number of generated plaintexts
    :param random_rows: indices in the range [0, 4) of rows to randomize
    :param random_cols: indices in the range [0, 4) of columns to randomize
    :param random_bytes: specific bytes to randomize. You should pass state
        bytes to randomize.
    :param seed: seed for deterministic generation of inputs.
    :param fixed_value: which value is written in fixed part of the plaintext.
    :returns: a Numpy array of shape ``(n_traces, 16)`` and dtype ``np.uint8``.

    """
    if random_cols is None:
        random_cols = []
    if random_rows is None:
        random_rows = []
    if random_bytes is None:
        random_bytes = []
    if seed is None:
        seed = random.randint(0, 2**64)

    assert all(0 <= x < 4 for x in random_rows), "row indices are in the range [0; 4)"
    assert all(0 <= x < 4 for x in random_cols), (
        "column indices are in the range [0; 4)"
    )
    assert all(0 <= x < 16 for x in random_bytes), (
        "state indices are in the range [0; 16)"
    )

    rng = np.random.default_rng(seed)

    # Generate an initial array.
    pts = np.full((n_traces, 16), fixed_value, dtype=np.uint8)

    # Randomize the parts requested by the user.
    for row in random_rows:
        pts[:, np.arange(0, 16, 4) + row] = rng.integers(
            0, 256, size=(n_traces, 4), dtype=np.uint8
        )
    for col in random_cols:
        pts[:, np.arange(4) + 4 * col] = rng.integers(
            0, 256, size=(n_traces, 4), dtype=np.uint8
        )
    for idx in random_bytes:
        pts[:, idx] = rng.integers(0, 256, n_traces, dtype=np.uint8)

    return pts


def biased_state_plaintexts(
    n_traces: int,
    key: np.ndarray,
    target_round: int,
    target_op: Union[str, AesOps],
    random_bytes: List[int] = None,
    seed: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generate plaintexts that induce bias state in selected round of a AES execution.

    The state at the target round (and target operation) is set to either
    ``[0x00, ..., 0x00]`` or ``[0xFF, ..., 0xFF]``. Then, the selected bytes (``random_bytes``)
    are randomized. Finally, the state is inverted to compute the input plaintext.

    :param key: AES key (128 bits, 16 bytes).
    :param target_round: AES round of the targeted aes operation
    :param target_op: targeted aes operation.
    :param n_traces: number of generated plaintexts
    :param random_bytes: unbiased bytes of the biased state (random)
    :param seed: seed for deterministic generation of inputs.

    :returns: A tuple of Numpy arrays, ``(labels, plaintexts)``. ``labels``has shape (``n_traces``) and
        contains 0 or 1. This indicates whether the state was filled with zeros or ones.
        The second member is ``plaintext``, and has shape ``(n_traces, 16)``.

    """

    if random_bytes is None:
        random_bytes = []
    if seed is None:
        seed = random.randint(0, 2**64)
    if isinstance(target_op, str):
        target_op = getattr(AesOps, target_op)
    rng = np.random.default_rng(seed)

    # Generate biased states
    labels = rng.integers(0, 2, size=n_traces, dtype=np.uint8)
    state = np.repeat(0xFF * labels[:, np.newaxis], 16, axis=1)
    for byte in random_bytes:
        state[:, byte] = np.random.randint(0, 256, size=n_traces, dtype=np.uint8)

    # Decrypt from intermediate round.
    cipher = AES(np.array(key, dtype=np.uint8))
    pts = cipher.decrypt(
        state,
        start_after=target_op,
        start_round=target_round,
    )

    return labels, pts


def generate_round_states(
    key: np.ndarray,
    pts: np.ndarray,
    target_op: Union[str, AesOps] = AesOps.add_round_key,
    model: str = "hw",
) -> np.ndarray:
    """
    Compute intermediate states of AES computation for a given operation.

    Distances on the first round are done with plaintexts.

    :param key: AES key (128 bits)
    :param pts: AES plaintexts
    :param target_op: AES operation that output the labels
    :param model: model applied on states, 3 supported model "hw" (identity), "hd" (xor) and "shd" (concatenation).

    """
    if isinstance(target_op, str):
        target_op = getattr(AesOps, target_op)

    cipher = AES(key)
    prev_state = pts.copy()

    all_labels = []
    for aes_round in range(1, 10):
        if aes_round == 1:
            state = cipher.encrypt(pts, stop_after=target_op, stop_round=1)
        else:
            state = cipher.encrypt(
                state,
                start_after=target_op,
                start_round=aes_round - 1,
                stop_after=target_op,
                stop_round=aes_round,
            )

        if model == "hw":
            label = state
        elif model == "hd":
            label = state ^ prev_state
        elif model == "shd":
            label = np.stack((prev_state, state), 0)
        else:
            raise ValueError(f"unsupported model: {model}")

        prev_state[:] = state[:]
        all_labels.append(label)

    return np.array(all_labels, dtype=np.uint8)


def _next_row(bts, row, kr, out_bias=255):
    idxs = np.roll(np.arange(4), -1 - row)[1:]
    out = out_bias
    out ^= AES_SBOX_TABLE[bts[..., idxs[0]]]
    out ^= AES_SBOX_TABLE[bts[..., idxs[1]]]
    out ^= GF_MULT2[AES_SBOX_TABLE[bts[..., idxs[2]]]]
    out ^= kr
    out = AES_INV_SBOX_TABLE[GF_INV_MULT3[out]]
    return out


def _gen_hd_bias(n, rk, biased_row, in_bias=0, out_bias=255):
    sr_matrix = AesOps.shift_rows(np.arange(16).reshape(1, 4, 4))[0].reshape(4, 4)
    base = np.random.randint(0, 256, size=(n, 16), dtype=np.uint8)

    for c in range(4):
        base[..., (c * 4 + 5 * biased_row) % 16] = in_bias
        last = _next_row(
            base[..., sr_matrix[c]],
            biased_row,
            rk[c * 4 + biased_row],
            out_bias=copy.copy(out_bias),
        )
        base[..., sr_matrix[c, (biased_row + 1) % 4]] = last
    return base


def _complete_column(column, column_key_round, label):
    """
    Complete the two last elements of an AES state column to generate aes state with HD bias on two rows.
    """
    column[..., 2] = GF_MULT3[AES_SBOX_TABLE[column[..., 0]]]
    column[..., 2] ^= AES_SBOX_TABLE[column[..., 1]]
    column[..., 2] ^= column[..., 0]
    column[..., 2] ^= column[..., 1]
    column[..., 2] ^= column_key_round[1] ^ column_key_round[0]
    column[..., 2] = GF_INV_MULT2[column[..., 2]]

    column[..., 3] = column[..., 2]
    column[..., 3] ^= GF_MULT2[AES_SBOX_TABLE[column[..., 0]]]
    column[..., 3] ^= GF_MULT3[AES_SBOX_TABLE[column[..., 1]]]
    column[..., 3] ^= column[..., 0]
    column[..., 3] ^= 0xFF * label
    column[..., 3] ^= column_key_round[0]

    column[..., 2] = AES_INV_SBOX_TABLE[column[..., 2]]
    column[..., 3] = AES_INV_SBOX_TABLE[column[..., 3]]


def biased_hd_plaintexts(
    n_traces: int,
    key: np.ndarray,
    target_round: int,
    biased_row: int = 0,
    sign: str = "unsigned",
    seed: Optional[int] = None,
):
    """
    Generate plaintexts that induce biased distance between two successive add round key outputs.

    The bias generated (0 or 0xFF) is available in labels returned.
    One state row (4 bytes) is biased.

    :param key: aes key (128 bits)
    :param target_round: aes round of the second add rond key
    :param n_traces: number of generated plaintexts
    :param biased_row: biased state row if 12 the two first rows are biased
    :param sign: sign of the distance 'unsigned', 'positive' or 'negative' (default:'unsigned')

    :returns: A tuple of Numpy arrays, ``(labels, plaintexts)``. ``labels``has shape (``n_traces``) and
      contains 0 or 1. This indicates whether the state was filled with zeros or ones.
      The second member is ``plaintext``, and has shape ``(n_traces, 16)``.

    """
    if seed is None:
        seed = random.randint(0, 2**64)
    rng = np.random.default_rng(seed)
    labels = rng.integers(0, 2, size=n_traces, dtype=np.uint8)

    # generate states that have biased hamming distances
    if sign == "unsigned":
        input_bias = rng.integers(0, 256, size=n_traces, dtype=np.uint8)
        output_bias = input_bias ^ labels * 0xFF
    elif sign == "positive":
        input_bias = np.zeros(n_traces, dtype=np.uint8)
        output_bias = labels * 0xFF
    elif sign == "negative":
        input_bias = labels * 0xFF
        output_bias = np.zeros(n_traces, dtype=np.uint8)
    else:
        raise ValueError(f"unsupported sign: {sign}")

    # get aes instance and round key
    cipher = AES(np.array(key, np.uint8))
    round_key = cipher._round_keys[target_round]

    if biased_row < 4:
        # generate biased states
        state = _gen_hd_bias(
            n_traces,
            round_key,
            biased_row,
            input_bias,
            output_bias,
        )
    elif biased_row == 12:
        # FIXME: we should have a different encoding for this.
        # generate biased states that induce 2 biased rows
        # with the 2 full rows, the entropy drops to 0
        state = input_bias.repeat(16).reshape(n_traces, 16)
        for col in range(4):
            # fmt: off
            _complete_column(
                state[..., col * 4:col * 4 + 4],
                round_key[col * 4:col * 4 + 4],
                labels,
            )
            # fmt: on
        state = AesOps.inv_shift_rows(state)
    else:
        raise ValueError(f"unsupported biased row: {biased_row}")

    # decrypt from input state to generate pts
    pts = cipher.decrypt(
        state,
        start_after=AesOps.inv_sub_bytes,
        start_round=target_round,
    )

    return labels, pts